''' local_descriptor.py
IR System for local descriptor
'''

import os, sys, time
import random

import numpy as np

__DEBUG__ = True

class LocalDescriptorIRHelper():
    def __init__(self):
        print('initialize LocalDescriptorIR ...')
        
    def __validate_format__(self, sample):
        lkeys = sample.keys()
        assert 'location_np_list' in lkeys, "'location_np_list' is not found."
        assert 'descriptor_np_list' in lkeys, "'descriptor_np_list' is not found."
        assert 'unique_id' in lkeys, "'unique_id' is not found."
    
    def __validate_format__(self, sample):
        lkeys = sample.keys()
        assert 'location_np_list' in lkeys, "'location_np_list' is not found."
        assert 'descriptor_np_list' in lkeys, "'descriptor_np_list' is not found."
        assert 'unique_id' in lkeys, "'unique_id' is not found."
        raise NotImplementedError()
   

    # public functios
    def build_index(
        self,
        x,
        m,
        n_bits):
        '''
        Args: 
            x: numpy array of descriptors, (ndescriptors x ndims)
            m: number of subquantizers
            n_bits: bits allocated per subquantizer
        Returns:
            indexed faiss
        '''
        dims = x.shape[1]
        
        # create index
        self.index = faiss.IndexPQ(dims, m, n_bits)
        
        # train
        self.index.train(x)
        
        if __DEBUG__:
            print('Training K-means for PQ ...')
            print('created index: ', self.index)

    def add_index(self, x_d):
        ''' populate the database
        '''
        self.index.add(x_d)

    def search(
        self,
        q_id,
        top_k = 10,
        method = 'BRUTE_FORCE'):
        '''
        Args:
            top_k: Top K index
        Returns:
            ret: id list of retrieved candidates
        '''
        assert method.upper() in ['BRUTE_FORCE', 'BF', 'PQ']
        raise NotImplementedError()


if __name__ == '__main__':
    pass;






















''' global_descriptor.py
IR System for local descriptor
'''

import os
import sys
import time
import random

import itertools
import abc
import numpy as np
from collections import Counter
from tqdm import tqdm
import concurrent.futures
from concurrent.futures import ThreadPoolExecutor

from irsys.feature_store import FeatureStore

__VERBOSE__ = True

class BaseSearch(metaclass=abc.ABCMeta):
    def __init__(
        self,
        with_mtc,
        qk,
        dk):
        self.fs = dict()
        self.fs['query'] = FeatureStore(with_mtc=with_mtc, qk=qk, dk=dk)
        self.fs['index'] = FeatureStore(with_mtc=with_mtc, qk=qk, dk=dk)
        self.qk = qk
        self.dk = dk
        
    def feed_from_batch(
        self,
        target,
        feat_list,
        unique_id_list):
        '''feed features to search system.
        index_feats: list of numpy array of index features, [np_arr1, np_arr2, np_arr3, ...]
        query_feats: list of numpy array of query features, [np_arr1, np_arr2, np_arr3, ...]
        '''
        assert isinstance(feat_list, list), 'type(index_feats) is not List.'
        if feat_list is None:
            return;
        tmp = [ self.fs[target].insert([ unique_id_list[k], feat_list[k] ], target) \
              for k, _ in enumerate(tqdm(feat_list)) ]
    
    def feed_from_single(
        self,
        target,
        feat,
        unique_id):
        '''feed feature one by one.
        '''
        if feat is None:
            return;
        self.fs[target].insert([unique_id, feat], target)
    
    @abc.abstractmethod
    def search_all(
        self,
        top_k,
        scoring,
        weights):
        '''get BF result from all given query ids.
        This requires excessive memory.
        Make sure top_k > kappa
        Returns:
            retrieved index indices sorted by similiarity.
        '''
    
    @abc.abstractmethod
    def search(
        self,
        query_feat,
        top_k,
        scoring,
        weights):
        '''
        Note that search one by one in BF is not efficient.
        '''

# Brute Force Search
class bfSearch(BaseSearch):
    def __init__(
        self,
        with_mtc,
        qk, dk):
        super(bfSearch, self).__init__(with_mtc, qk, dk)
    
    def __get_maiv_terms__(self, x, kappa):
        '''get maiv terms
        return maximally activated index vector(maiv)
        '''
        return np.argsort(x)[::-1][:kappa]

    def __calc_match_term_count__(self, x1, x2):
        return len(np.intersect1d(x1, x2, assume_unique=True).tolist())
 
    def search_all(
        self,
        top_k,
        score_method,
        weights):
        if score_method == 'cosim':
            query_feats = self.fs['query'].export(mtc=False)
            index_feats = self.fs['index'].export(mtc=False)
            if isinstance(query_feats, list):
                query_feats = np.asarray(query_feats)
            if isinstance(index_feats, list):
                index_feats = np.asarray(index_feats)
            assert query_feats.shape[1] == index_feats.shape[1], \
                'feature dimension of query: {}, index:{} does not match' \
                .format(index_feats.shape[1], query_feats.shape[1])
            # https://oss.navercorp.com/VisualSearch/simage-qlty/blob/master/evaluation.py
            # mAP could be differerent depending on the use of np.argsort() or python sort().
            # To Do: Check if the result of python sort() and np.argsort() is same or not.
            print('calculating cosine similarity score...')
            cosim = np.dot(query_feats, index_feats.T)
            return np.argsort(-cosim, axis=1)[:,:top_k] \
                   if top_k is not None \
                   else np.argsort(-cosim, axis=1)

        elif score_method == 'mtc':
            q_feats = self.fs['query'].export(mtc=True)
            i_feats = self.fs['index'].export(mtc=True)
            if isinstance(q_feats, list):
                q_feats = np.asarray(q_feats)
            if isinstance(i_feats, list):
                i_feats = np.asarray(i_feats)
            mtc = np.dot(q_feats, i_feats.T)
            return np.argsort(-mtc, axis=1)[:,:top_k] \
                   if top_k is not None \
                   else np.argsort(-mtc, axis=1)

    def search(
        self,
        query_feat,
        top_k,
        score_method,
        weights):
        if score_method == 'cosim':
            index_feats = self.fs['index'].export(mtc=False)
            if isinstance(query_feat, list):
                query_feats = np.asarray(query_feats)
            if isinstance(index_feats, list):
                index_feats = np.asarray(index_feats)
            cosim = np.dot(query_feat, index_feats.T)
            return np.argsort(-cosim, axis=0)[:top_k] \
                   if top_k is not None \
                   else np.argsort(-cosim, axis=0)

        if score_method == 'mtc':
            i_feats = self.fs['index'].export(mtc=True)
            q_feat = self.fs['query'].__get_maiv_terms__(query_feat, self.qk)
            if isinstance(q_feat, list):
                q_feat = np.asarray(q_feat)
            if isinstance(i_feats, list):
                i_feats = np.asarray(i_feats)
            mtc = np.dot(q_feat, i_feats.T)
            return np.argsort(-mtc, axis=0)[:top_k] \
                   if top_k is not None \
                   else np.argsort(-mtc, axis=0)

class fashionSearch(BaseSearch):
    '''This is special search class only for fashion.
    '''
    def __init__(
        self,
        with_mtc,
        qk, dk):
        super(fashionSearch, self).__init__(with_mtc, qk, dk)
 
    def search_all(
        self,
        top_k,
        score_method,
        weights):
        assert (isinstance(weights, tuple) or isinstance(weights, list)), \
               'Invalid argument weights: {}'.format(type(weights))
        if score_method == 'cosim':
            query_feats = self.fs['query'].export(mtc=False)
            index_feats = self.fs['index'].export(mtc=False)
            # get weighted cosine similarity for shape/texture/color
            cosim = None
            for i, w in enumerate(weights):
                q_feats = [ x[i] for x in query_feats ]
                i_feats = [ x[i] for x in index_feats ]
                if isinstance(query_feats, list):
                    q_feats = np.asarray(q_feats)
                if isinstance(index_feats, list):
                    i_feats = np.asarray(i_feats)
                assert q_feats.shape[1] == i_feats.shape[1], \
                    'feature dimension of query: {}, index:{} does not match' \
                    .format(i_feats.shape[1], q_feats.shape[1])
                if i == 0:
                    cosim = w * np.dot(q_feats, i_feats.T)
                else:
                    cosim = cosim + float(w) * np.dot(q_feats, i_feats.T)

            return np.argsort(-cosim, axis=1)[:,:top_k] \
                   if top_k is not None \
                   else np.argsort(-mtc, axis=1)
        
        elif score_method == 'mtc':
            query_feats = self.fs['query'].export(mtc=True)
            index_feats = self.fs['index'].export(mtc=True)
            mtc = None
            for i, w in enumerate(weights):
                q_feats = [ x[i] for x in query_feats ]
                i_feats = [ x[i] for x in index_feats ]
                if isinstance(q_feats, list):
                    q_feats = np.asarray(q_feats)
                if isinstance(i_feats, list):
                    i_feats = np.asarray(i_feats)
               
                tmp = np.dot(q_feats, i_feats.T)
                mtc = tmp if i == 0 else mtc + float(w)*tmp
            # sorting
            if __VERBOSE__:
                print('sorting by score...')
            return np.argsort(-mtc, axis=1)[:,:top_k] \
                   if top_k is not None \
                   else np.argsort(-mtc, axis=1)
            
    def search(
        self,
        query_feat,
        top_k,
        score_method,
        weights):
        assert (isinstance(weights, tuple) or isinstance(weights, list)), \
               'Invalid argument weights: {}'.format(type(weights))
        if score_method == 'cosim':
            raise NotImplementedError()
        if score_method == 'mtc':
            raise NotImplementedError()
        

# sklearn Search
class sklearnSearch(BaseSearch):
    def __init__(self):
        raise NotImplementedError()
    
    def search(self):
        raise NotImplementedError()

# Flann Search
class flannSearch(BaseSearch):
    def __init__(self):
        raise NotImplementedError()
    
    def search(self):
        raise NotImplementedError()

# IR Helper
class GlobalDescriptorIRHelper():
    def __init__(
        self,
        srch_lib,
        srch_method,
        with_mtc,
        qk=30, dk=120,
        verbose=False):
        '''GlobalDescriptorIRHelper
        '''
        # select ir system method.
        self.srch_lib = srch_lib
        self.srch_method = srch_method.lower()
        self.with_mtc = with_mtc
        self.qk = qk
        self.dk = dk
        __VERBOSE__ = verbose
        self.__feature_path__ = dict()
        if self.srch_method == 'bf':
            if self.srch_lib is None:
                self.irsys = bfSearch(
                    with_mtc=with_mtc,
                    qk=qk, dk=dk)
        elif self.srch_method == 'fashion':
            self.irsys = fashionSearch(
                with_mtc=with_mtc,
                qk=qk, dk=dk)
        else:
            raise ValueError('Invalid IR method {}!!!'.format(method))

    def __load_feature_from_path__(
        self,
        target,
        unique_id_list):
        '''load features from given unique_id list.
        '''
        assert target in self.__feature_path__, \
            'feed {} before load features.'.format(target)
        feat_list = [np.load(os.path.join(self.__feature_path__[target], x+'.npy'))
                    for x in unique_id_list]
        assert len(unique_id_list) == len(feat_list), \
            'error: len(unique_id_list) != len(feat_list)'
        return feat_list, unique_id_list

    def feed_from_path(
        self,
        path,
        target,
        unique_id_list):
        '''feed feaures to irsystem with unique id of images
        '''
        assert isinstance(unique_id_list, list), \
            'type(unique_id_list) is not List.'
        assert target in ['index', 'query'], \
            '[feed_from_path] Invalid target argument: {}'.format(target)
      
        self.__feature_path__[target] = path
        print('Loading {} features({}) from {}'\
            .format(target, len(unique_id_list), path))
        
        # update features.
        if __VERBOSE__:
            print('Feed batch {} features to {} IR System'.format(target, self.srch_method))
        feat_list, unique_id_list = self.__load_feature_from_path__(
            target=target,
            unique_id_list=unique_id_list)
        self.irsys.feed_from_batch(
            target=target,
            unique_id_list=unique_id_list,
            feat_list=feat_list)
        return len(feat_list)

    def feed(
        self,
        target,
        feat,
        unique_id):
        '''feed features to irsystem directly with feat and unique_id.
        '''
        # update features.
        #if __VERBOSE__:
        #    msg = 'Feed single {}({}) feature to {} IR System'
        #    print(msg.format(target, unique_id, self.srch_method))
        if target == 'query':
            kappa = self.qk
        else:
            kappa = self.dk

        self.irsys.feed_from_single(
            target=target,
            feat=feat,
            unique_id = unique_id)

    def load_feat_by_unique_id(
        self,
        unique_id,
        target,
        mtc=False):
        '''load single feature with unique id.
        '''
        assert target in ['query', 'index']
        assert self.irsys.fs[target].exist(unique_id), \
            'unique id {} not found. please do "feed" before "load"'.format(unique_id)
        idx = self.id_to_idx_fn(unique_id, target)
        return self.irsys.fs[target].read(idx)
    
    def load_feat_by_index(
        self,
        idx,
        target,
        mtc=False):
        '''load single feature with index
        '''
        assert target in ['query', 'index']
        return self.irsys.fs[target].read(idx, mtc)

    def search_all(
        self,
        top_k,
        score_method,
        weights=None):
        '''search with all given queries
        '''
        return self.irsys.search_all(
            top_k,
            score_method,
            weights)

    def search_by_unique_id(
        self,
        query_id,
        top_k,
        score_method,
        weights=None):
        '''search with single query id.
        '''
        qidx = self.id_to_idx_fn(query_id, 'query')
        if score_method=='mtc':
            q_feat = self.irsys.fs['query'].read(qidx, mtc=True)
        else:
            q_feat = self.irsys.fs['query'].read(qidx, mtc=False)
        return self.irsys.search(
            q_feat,
            top_k,
            score_method,
            weights)

    def search_by_feat(
        self,
        feat,
        top_k,
        score_method,
        weights=None):
        '''search with the feature directly.
        '''
        return self.irsys.search(
            feat,
            top_k,
            score_method,
            weights)

    def sample(
        self,
        target,
        idx,
        mtc):
        assert target in ['index', 'query'], '[sample] Invalid target argument: {}'.format(target)
        if idx == -1:
            idx = random.randrange(0, len(self.irsys.fs[target].__raw_features__))
        return (self.irsys.fs[target].read(idx, mtc=mtc),
                self.idx_to_id_fn(idx, target), idx)
       
    def idx_to_id_fn(
        self,
        idx,
        target):
        return self.irsys.fs[target].idx_to_id(idx)
    
    def id_to_idx_fn(
        self,
        _id,
        target):
        return self.irsys.fs[target].id_to_idx(_id)
    
    
    def get_data_length(self, target):
        return len(self.irsys.fs[target].__raw_features__)

if __name__ == '__main__':
    pass;
























import concurrent.futures
from concurrent.futures import ThreadPoolExecutor

import utils.utils as utils
from evals.evaluate import (PureInstanceRetrieval,
                            AttributeManipulation)

__DEBUG__ = True

class EvalHelper(object):
    def __init__(
        self):
        self.rank_dict = dict()
        self.gt_dict = dict()
        pass;
    
    def feed_gt_from_path(
        self,
        path):
        print('load json file from {}'.format(path))
        gt_dict = utils.load_dict_from_path(path)
        self.feed_gt_from_dict(gt_dict)
    
    def feed_rank_from_path(
        self,
        path):
        print('load json file from {}'.format(path))
        rank_dict = utils.load_dict_from_path(path)
        self.feed_rank_from_dict(rank_dict)

    def feed_gt_from_dict(
        self,
        gt_dict):
        assert isinstance(gt_dict, dict), \
            'Invalid gt_dict argument type: {}'.format(type(gt_dict))
        try:
            for k, v in gt_dict.items():
                if not 'positive' in v:
                    raise ValueError('no "positive" field in gt data.')
                break;
            self.gt_dict = gt_dict
        except Exception as e:
            raise ValueError(e)
    
    def feed_rank_from_dict(
        self,
        rank_dict):
        assert isinstance(rank_dict, dict), \
            'Invalid rank_dict argument type: {}'.format(type(rank_dict))
        try:
            for k, v in rank_dict.items():
                if not isinstance(v, list):
                    raise ValueError('Invalid rank dictionary format')
            self.rank_dict = rank_dict
        except Exception as e:
            raise ValueError(e)

    def feed_gt(
        self,
        data):
        '''
        ('<q_id>', [<i_id1>, <i_id2>, ...])
        '''
        assert (isinstance(data, list) or isinstance(data, tuple)), \
            'Argument type data must be tuple or list: {}'.format(type(data))
        new_data = {
            data[0]: {
                'positive': data[1],
                'junk': None
            }
        }
        self.gt_dict.update(new_data)

    def feed_rank(
        self,
        data):
        '''
        ('<q_id>', [<i_id1>, <i_id2>, ...])
        '''
        assert (isinstance(data, list) or isinstance(data, tuple)), \
            'Argument type data must be tuple or list: {}'.format(type(data))
        new_data = {
            data[0]: data[1]
        }
        self.rank_dict.update(new_data)

    def evaluate(
        self,
        metric=['mAP'],
        kappa=None,
        indiv=False):
        '''evaluate
        metric: mAP / Precision / Recall / Top_k_accuracy
        indiv: True will return metric result per each individual query.
        '''
        if __DEBUG__:
            print('--------------------- config ----------------------')
            print('metric: {}'.format(metric))
            print('kappa: {}'.format(kappa))
            print('indiv: {}'.format(indiv))
            print('---------------------------------------------------')
       
        if isinstance(metric, str):
            metric = [ metric ]
        if (kappa is None) or (isinstance(kappa, int)):
            kappa = [ kappa ]
        assert isinstance(kappa, list), \
            'Invalid argument type kappa: {}'.format(type(kappa))
        
        # multi-threading.
        submit_list = []
        thread_pool = ThreadPoolExecutor(max_workers = min(len(kappa), 10))
       
        ret = []
        for m in metric:
            if m.lower() == 'map':
                for k in kappa:
                    s = thread_pool.submit(
                        PureInstanceRetrieval.compute_mAP,
                        self.rank_dict,
                        self.gt_dict,
                        indiv,
                        k)
                    submit_list.append(s)
                
                result = dict()
                completed = concurrent.futures.as_completed(submit_list)
                for x in completed:
                    res = x.result()
                    result.update({str(res[1]):res[0]})
                ret.append(result)
            
            elif m.lower() == 'top_k_acc':
                for k in kappa:
                    s = thread_pool.submit(
                        PureInstanceRetrieval.compute_top_k_acc,
                        self.rank_dict,
                        self.gt_dict,
                        k)
                    submit_list.append(s)
                
                result = dict()
                completed = concurrent.futures.as_completed(submit_list)
                for x in completed:
                    res = x.result()
                    result.update({str(res[1]):res[0]})
                ret.append(result)

        return ret
        
        



'''evaluate.py
evalutate mAP
'''

import csv
import json

import tqdm

__DEBUG__ = False



class PureInstanceRetrieval(object):
    '''For Instance Retrieval
    '''
    @staticmethod
    def compute_AP(
        pos_list,
        junk_list,
        rank_list,
        kappa):
        '''
        Ignore junk images if exists.
        reference: https://sites.google.com/site/hyunguk1986/personal-study/-ap-map-recall-precision
        reference: http://www.robots.ox.ac.uk/~vgg/data/oxbuildings/compute_ap.cpp
        '''
        num_pos_list = len(pos_list)
        if kappa is not None:
            rank_list = rank_list[:kappa]
            num_pos_list = min(num_pos_list, kappa)
    
        old_recall = 0.
        old_precision = 1.
        ap = 0.
    
        intersect_size = 0
        j = 0
        for i, rank in enumerate(rank_list):
            if rank in junk_list:
                continue;
            if rank in pos_list:
                intersect_size += 1
            
            j += 1.
            recall = intersect_size / float(num_pos_list)
            precision = intersect_size / float(j)
    
            ap += (recall - old_recall) * ((old_precision + precision)/2.)      # area of trapezoidal.
            #ap += (recall - old_recall) * (precision)                          # area of outer rectangle
       
            # udpate recall and precision.
            old_recall = recall
            old_precision = precision
            
            # escape if recall == 1.0
            if intersect_size >= num_pos_list:
                break;
    
        return ap, recall, precision
    
    
    @staticmethod
    def compute_mAP(
        rank_dict,
        gt_dict,
        indiv = True,
        kappa = None):
        '''compute mAP given set of return results.
        Args:
            gt_dict: Ground Truth json object.
            {
                <query_unique_id>: {
                    'positive': [ pos1, pos2, pos3, ... ],
                    'junk': [ junk1, junk2, junk3, ... ]
            }
            rank_dict: Prediction json object.
            {
                <query_unique_id>: [ rank1, rank2, rank3, ... ]
            }
        '''
        result = dict()
        result['overall'] = dict()
        if indiv:
            result['indiv'] = dict()
    
        mAP = 0.
        Precision = 0.
        Recall = 0.
        ncnt = 0
                
        for uid, rank_list in rank_dict.items():
            if not uid in gt_dict:
                print('no {} key exists in gt_dict'.format(uid))
                continue;   # if id not exists in GT, skip.
            pos_list =gt_dict[uid]['positive']
            junk_list = []       # ignore junk.
                
            # if GT list is empty, skip.
            if len(pos_list) > 0:
                ap, recall, precision = PureInstanceRetrieval.compute_AP(
                    pos_list,
                    junk_list,
                    rank_list,
                    kappa)
                
                mAP = mAP + ap
                Recall = Recall + recall
                Precision = Precision + precision
                ncnt = ncnt + 1
                
                if indiv:
                    if not uid in result['indiv']:
                        result['indiv'][uid] = dict()
                    data = dict()
                    data['ap'] = ap
                    data['recall'] = recall
                    data['precision'] = precision
                    result['indiv'][uid].update(data)
            else:
                print('no ground-truth was found for query: {}'.format(uid))
        
        
        # calc. ooveralll 
        kappa = 'All' if kappa is None else kappa
        mAP = mAP / float(ncnt)
        Precision = Precision / float(ncnt)
        Recall = Recall / float(ncnt)
        result['overall']['mAP'] = mAP
        result['overall']['recall'] = Recall
        result['overall']['precision'] = Precision
    
        return result, kappa
    
    @staticmethod
    def compute_top_k_acc(
        rank_dict,
        gt_dict,
        kappa=None):
        """compute top-k retrieval accuracy.
        """
        result = dict()
        hit = 0
        ncnt = 0
        for uid, rank_list in rank_dict.items():
            if not uid in gt_dict:
                print('no {} key exists in gt_dict'.format(uid))
                continue;   # if id not exists in GT, skip.
            pos_list =gt_dict[uid]['positive']
            junk_list = []       # ignore junk.
            
            # if GT list is empty, skip.
            if len(pos_list) > 0:
                rank_list = rank_list[:kappa] if kappa is not None else rank_list
                # hit if at least one gt exists in rank list.
                inter_sect = list(set(rank_list).intersection(pos_list))
                ncnt = ncnt + 1
                if len(inter_sect) > 0:
                    hit = hit + 1

        kappa = 'All' if kappa is None else kappa
        top_k_acc = hit / float(ncnt)
        result['top_k_acc'] = top_k_acc
        return result, kappa 

class AttributeManipulation(object):
    ''' For Attribute Mainpulation
    ref: Memory-Augmented Attribute Manipulation Networks for Interactive Fashion Search
    '''
    @staticmethod
    def compute_NDCG(
        inv_gt_json,
        rank_json,
        kappa=5):
        '''
        NDCG@k = \sum_{j=1}^{k}\frac{ 2^{rel(j)-1} }{log(j+1)}
        
        inv_gt_json: {
            <query_unique_id>: {
                'positive': [attr1, attr2, attr3]
            }
        }
        '''
        for targ_attr, res_dict in rank_json.items():
            for q_id, rank_dict in res_dict.items():
                for attr_key, rank_list in rank_dict.items():
                    assert attr_key in gt_json, 'attr_key "{}" does not exist in gt_json!!!'.format(attr_key)
                    gt_list = gt_json[attr_key]['positive']
                    for r_id in rank_list:
                       pass; 
    
    @staticmethod
    def compute_mAP(
        gt_json,
        rank_json,
        kappa=5):
        """compute top-k retrieval accuracy.
        we consider a hit if the method finds
        at least one clothing image with exactly
        the same attributes as indicated by the query 
        in top-k result. 
        
        Args:
            gt_json: Ground Truth json object.
            {
                <attr_comb_key>: {
                    'positive': [ pos1, pos2, pos3, ... ],
                }
            }
            rank_json: Prediction json object.
            {
                <targ_attr>: {
                    <query_unique_id>: {
                        <attr_comb_key1>: [rank1, rank2, rank3, ...],
                        <attr_comb_key2>: [rank1, rank2, rank3, ...],
                        <attr_comb_key3>: [rank1, rank2, rank3, ...],
                        <attr_comb_key4>: [rank1, rank2, rank3, ...],
                        <attr_comb_key5>: [rank1, rank2, rank3, ...],
                    }
                }
            }
        """
        res_map = {}
        all_cnt = 0
        all_mAP = 0.
        qq = []
        ss = []
        for targ_attr, res_dict in tqdm.tqdm(rank_json.items()):
            mAP = 0.
            cnt = 0
            for q_id, rank_dict in tqdm.tqdm(res_dict.items()):
                for attr_key, rank_list in rank_dict.items():
                    if attr_key in gt_json:
                        gt_list = gt_json[attr_key]['positive']
                        ap = compute_AP(gt_list, [], rank_list, kappa=kappa)
                        mAP = mAP + ap
                        cnt = cnt + 1
                    else:       # skip if attr_key does not exists in gt.
                        continue;
            all_mAP = all_mAP + mAP
            all_cnt = all_cnt + cnt
            mAP = mAP / cnt
            res_map[targ_attr] = mAP
        res_map['all'] = all_mAP / all_cnt
        return res_map          
    
    @staticmethod
    def compute_top_k_acc(
        gt_json,
        rank_json,
        kappa=5):
        """compute top-k retrieval accuracy.
        we consider a hit if the method finds
        at least one clothing image with exactly
        the same attributes as indicated by the query 
        in top-k result. 
        
        Args:
            gt_json: Ground Truth json object.
            {
                <attr_comb_key>: {
                    'positive': [ pos1, pos2, pos3, ... ],
                }
            }
            rank_json: Prediction json object.
            {
                <targ_attr>: {
                    <query_unique_id>: {
                        <attr_comb_key1>: [rank1, rank2, rank3, ...],
                        <attr_comb_key2>: [rank1, rank2, rank3, ...],
                        <attr_comb_key3>: [rank1, rank2, rank3, ...],
                        <attr_comb_key4>: [rank1, rank2, rank3, ...],
                        <attr_comb_key5>: [rank1, rank2, rank3, ...],
                    }
                }
            }
        """
        res_acc = {}
        all_cnt = 0
        all_hit = 0.
        qq = []
        ss = []
        for targ_attr, res_dict in tqdm.tqdm(rank_json.items()):
            hit = 0.
            cnt = 0
            for q_id, rank_dict in tqdm.tqdm(res_dict.items()):
                for attr_key, rank_list in rank_dict.items():
                    if attr_key in gt_json:
                        gt_list = gt_json[attr_key]['positive']
                        inter_sect = list(set(rank_list[:kappa]).intersection(gt_list))
                        # hit if at least one attribute matched image is found.
                        cnt = cnt + 1
                        if len(inter_sect) > 0:
                            hit = hit + 1
                    else:       # skip if attr_key does not exists in gt.
                        continue;
            all_cnt = all_cnt + cnt
            all_hit = all_hit + hit
            res_acc[targ_attr] = hit/float(cnt)
        res_acc['all'] = all_hit / float(all_cnt)
        
        return res_acc


if __name__ == "__main__":
    gt_json = {
        'shirts.blue.checked': {
            'positive': ['a1','a2','a3','a4','a5','a6']
        },
        'shirts.yellow.checked': {
            'positive': ['b1','b2','b3','b4','b5','b6']
        },
        'shirts.green.checked': {
            'positive': ['c1','c2','c3','c4','c5','c6']
        },
        'coats.blue.checked': {
            'positive': ['d1','d2','d3','d4','d5','d6']
        },
        'onepiece.blue.checked': {
            'positive': ['e1','e2','e3','e4','e5','e6']
        },
        'shirts.blue.flower': {
            'positive': ['f1','f2','f3','f4','f5','f6']
        },
        'shirts.blue.dot': {
            'positive': ['g1','g2','g3','g4','g5','g6']
        },
    }
    rank_json = {
        'color': {
            'q1': {
                'shirts.blue.checked': ['a1','a3','b1','b8','c3','a2','h7'],
                'shirts.yellow.checked': ['g1','u8','b1','a7','j3','e4','p7'],
                'shirts.green.checked': ['a6','a4','c2','n6','d1','h2','a5'],
            }
        },
        'shape': {
            'q1': {
                'shirts.blue.checked': ['e1','a6','f1','b8','f3','a2','h7'],
                'coats.blue.checked': ['g1','u8','b1','d7','j3','d4','e7'],
                'onepiece.blue.checked': ['e6','e4','e2','n6','d1','h2','a5'],
            }
        },
        'pattern': {
            'q1': {
                'shirts.blue.checked': ['a1','a3','b1','b8','a3','a2','h7'],
                'shirts.blue.flower': ['f1','f2','b1','a7','j3','e4','p7'],
                'shirts.blue.dot': ['g5','a4','g2','n6','d1','h2','a5'],
            }
        },
    }
    acc = am_compute_mAP(
        gt_json,
        rank_json)
    print(acc) 






'''utils.py
'''
import os
import sys
import time
import random
import csv
import json
import glob

import pickle
import uuid
import numpy as np

__DEBUG__ = False

def mkdir_p(path):
    if not os.path.exists(path):
        os.makedirs(path)

def save_file(path, mode='npy'):
    assert mode in ['npy', 'pickle']
    if mode == 'npy':
        raise ValueError('numpy')
    elif mode == 'pickle':
        raise ValueError('pickle')

def load_pil_image(path):
    with open(path, 'rb') as f:
        return Image.open(f).convert('RGB')

def save_pil_image(path):
    raise NotImplementedError()

def visualize_ransac_match(img1, img2, kp1, kp2):
    raise NotImplementedError()

def visualize_correspondance_match(img1, img2, kp1, kp2):
    raise NotImplementedError()

def read_file(path):
    raise NotImplementedError()

def convert_indices_list_to_ids_list(indices_list, idx_to_iid_fn):
    '''convert indices_list to index_ids_list
    '''
    result_ids_list = []
    if len(indices_list.shape) == 1:
        indices_list = [ indices_list ]
    for i, indices in enumerate(indices_list):
        ids = [ idx_to_iid_fn(x) for x in indices ]
        t = [ x for x in indices ]
        result_ids_list.append(ids)
    return result_ids_list

def save_submission(
    path,
    filename,
    query_id_list,
    rank_ids_list,):
    '''save rank result to submission file format.
    '''
    # make sure path exist.
    mkdir_p(path)
   
    # save submission file.
    f = open(os.path.join(path, filename), 'w', encoding ='utf-8', newline='')
    wr = csv.writer(f, delimiter=' ', escapechar=' ', quoting=csv.QUOTE_NONE)
    for idx, rank_ids in enumerate(rank_ids_list):
        msg = []
        msg.append(query_id_list[idx])
        msg.append(','.join(rank_ids))
        wr.writerow(msg)
    f.close()
    print('saved submission file @ {}'.format(os.path.join(path, filename)))


def load_submission(path):
    '''return json object.
    Input file format:
        <query_unique_id1> <uid11>,<uid12>,<uid13>,...
        <query_unique_id2> <uid21>,<uid22>,<uid23>,...
        <query_unique_id3> <uid31>,<uid32>,<uid33>,...
    Return json format:
        {
            <query_unique_id1>: [ uid11, uid12, uid13, ... ]
            <query_unique_id2>: [ uid21, uid22, uid23, ... ]
            <query_unique_id3>: [ uid31, uid32, uid33, ... ]
        }
    '''
    json_list = {}
    with open(path) as f:
        for line in f:
            line = line.strip()
            tmp = line.split(' ')
            if len(tmp) > 1: # skip if no GT positive images found.
                q_id = tmp[0].strip()
                c_ids = tmp[1].strip().split(',')
                json_list[q_id] = c_ids
    return json_list

def load_unique_id_list_from_file(file_path):
    '''load_unique_id_list_from_file
    read list of query and index images form csv file.
    '''
    # query id list
    with open(file_path, 'r', encoding='utf-8') as f:
        rdr = csv.reader(f)
        unique_id_list = [ str(line[0].strip()) for line in rdr ]
        return unique_id_list

def load_unique_id_list_from_path(path):
    '''load_unique_id_list_from_path
    read list of query and index images form csv file.
    '''
    #assert ext is not None, 'ext is not specified'
    image_list = glob.glob(os.path.join(path, '*.npy'))
    unique_id_list = list(set([os.path.splitext(os.path.basename(x))[0] for x in image_list]))
    return unique_id_list

def load_dict_from_path(path):
    with open(path) as json_file:
        return json.load(json_file)
    
def generate_unique_id():
    return str(uuid.uuid4()).replace('-','')

def calculate_avg_prec(lgt, lpred, kappa):
    '''calculate AP given single query.
    Args:
        lgt: list of GT positive unique_ids
        lpred: list of retrieved unique_ids
    Returns:
        ap: average precision
    '''
    # evaluate top_k result only
    num_lgt = 0
    if kappa is None:
        num_lgt =len(lgt)
    else:
        num_lgt = min(kappa, len(lgt))

    ap = 0.
    lpred = lpred[:num_lgt]
    lcorrect = [ x for x in lpred if x in lgt ]
    ap = len(lcorrect) / float(num_lgt)

    if __DEBUG__:
        print('ap:{:.4f} %'.format(ap*100))

    return ap

if __name__ == '__main__':
    pass;
    
    




'''feature_store.py
'''

import numpy as np
from zlib import crc32

class FeatureStore(object):
    def __init__(
        self,
        with_mtc,
        qk, dk):
        '''FeatureStore
        '''
        self.__with_mtc__ = with_mtc
        self.qk = qk
        self.dk = dk
        self.clean()

    def __hash_str_to_float__(self, s, encoding="utf-8"):
        return 1e-3 * float(crc32(s.encode(encoding)) & 0xffffffff) / 2**32

    def __get_maiv_terms__(self, x, kappa, eps=0.0):
        '''get maiv terms
        return maximally activated index form the vector
        '''
        if isinstance(x, tuple):
            res = list()
            for i, _x in enumerate(x):
                t = np.zeros(_x.shape[0], dtype=np.int8)
                indices = np.argsort(_x)[::-1][:kappa]
                t[indices] = 1.0 + eps
                res.append(t)
            return tuple(res)
        else:
            t = np.zeros(x.shape[0], dtype=np.int8)
            indices = np.argsort(x)[::-1][:kappa]
            t[indices] = 1.0 + eps
            return t
   
    def clean(self):
        self.__idx_to_id__ = list()
        self.__id_to_idx__ = dict()
        self.__raw_features__ = list()
        self.__mtc_features__ = list()
        
    def insert(self, x, target):
        '''
        x = [unique_id, feat]
        '''
        assert isinstance(x, list) and len(x) == 2
        if target == 'query':
            kappa = self.qk
        elif target == 'index':
            kappa = self.dk
        
        uid, feat = x
        if not uid in self.__id_to_idx__:
            self.__id_to_idx__.update(
                { uid: len(self.__idx_to_id__) }
            )
            self.__idx_to_id__.append(uid)
            self.__raw_features__.append(feat)
            if self.__with_mtc__:
                self.__mtc_features__.append(self.__get_maiv_terms__(
                    feat,
                    kappa,
                    self.__hash_str_to_float__(uid)))
            return True
        else:
            print('duplicated unique id found: {}'.format(uid))
            return False

    def export(self, mtc=False):
        if mtc:
            return self.__mtc_features__
        else:
            return self.__raw_features__

    def read(self, idx, mtc=False):
        if mtc:
            return self.__mtc_fatures__[idx]
        else:
            return self.__raw_features__[idx]

    def idx_to_id(self, idx):
        return self.__idx_to_id__[idx]
    
    def id_to_idx(self, _id):
        return self.__id_to_idx__[_id]
    
    def delete(self):
        pass;

    def exist(self, uid):
        return uid in self.__id_to_idx__



"""
class FashionFeatureStore(FeatureStore):
    def __init__(
        self,
        with_mtc,
        qk, dk):
        super(FeatureStore, self).__init__(with_mtc, qk, dk)
       
    def clean(self):
        self.__idx_to_id__ = dict()
        self.__id_to_idx__ = dict()
        self.__raw_features__ = list()
        
        self.__raw_features__['shape'] = list()
        self.__raw_features__['texture'] = list()
        self.__raw_features__['color'] = list()
        self.__mtc_features__ = dict()
        self.__mtc_features__['shape'] = list()
        self.__mtc_features__['texture'] = list()
        self.__mtc_features__['color'] = list()
    
    def insert(self, x, target):
        '''
        x = [unique_id, (shape_feat, texture_feat, color_feat)]
        '''
        assert len(x) == 2
        assert isinstance(x[1], tuple) or isinstance(x[1], list)

        if target == 'query':
            kappa = self.qk
        elif target == 'index':
            kappa = self.dk
        uid, feat = x
        if not uid in self.__id_to_idx__:
            self.__id_to_idx__.update(
                { uid: len(self.__idx_to_id__) }
            )
            self.__idx_to_id__.append(uid)
            eslf.__raw_features_.append
            self.__raw_features__['shape'].append(feat[0])
            self.__raw_features__['texture'].append(feat[1])
            self.__raw_features__['color'].append(feat[2])
            if self.__with_mtc__:
                self.__mtc_features__['shape'].append(self.__get_maiv_terms__(feat[0], kappa))
                self.__mtc_features__['texture'].append(self.__get_maiv_terms__(feat[1], kappa))
                self.__mtc_features__['pattern'].append(self.__get_maiv_terms__(feat[2], kappa))
            return True
        else:
            print('duplicated unique id found: {}'.format(uid))
            return False
"""
    



from setuptools import setup, find_packages

setup(name='irbench-vision',
      version='0.1.9',
      description='Image Retrieval Performance Benchmark on Large-scale Dataset',
      url='',
      author='Minchul Shin',
      author_email='min.nashory@navercorp.com',
      license='MIT',
      packages=find_packages(),
      zip_safe=False)
